# Sage version information for shell scripts
# This file is auto-generated by the sage-update-version script, do not edit!
SAGE_VERSION='9.6.beta3'
SAGE_RELEASE_DATE='2022-02-27'
SAGE_VERSION_BANNER='SageMath version 9.6.beta3, Release Date: 2022-02-27'
