# Sage version information for Python scripts
# This file is auto-generated by the sage-update-version script, do not edit!
version = '9.6.beta3'
date = '2022-02-27'
banner = 'SageMath version 9.6.beta3, Release Date: 2022-02-27'
